"""
Fairness Oriented Multiobjective Optimization (Fomo)

BSD 3-Clause License

Copyright (c) 2023, William La Cava

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright notice, this
   list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright notice,
   this list of conditions and the following disclaimer in the documentation
   and/or other materials provided with the distribution.

3. Neither the name of the copyright holder nor the names of its
   contributors may be used to endorse or promote products derived from
   this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
"""
from sklearn.neural_network import MLPClassifier
from sklearn.utils import check_random_state
from sklearn.compose import ColumnTransformer
from sklearn.preprocessing import OneHotEncoder
# from itertools import chain
import numpy as np
import pandas as pd
from scipy.special import expit

class MLP(MLPClassifier):

    def get_n_weights(self):
        """Sets coefs_ and intercepts_ to x."""
        return np.sum([c.size for c in self.coefs_ + self.intercepts_])

    def set_weights(self, x):
        """Sets coefs_ and intercepts_ to x."""
        self._unpack(x)

    def fit(self, X, y):
        raise NotImplementedError('Dont call fit on this class!')
        return self
    
    def predict(self, X):
        # one-hot encode X
        # X = pd.get_dummies(X.astype('category'))
        return self.predict_proba(self._one_hot_encode(X))[:,1]

    def _one_hot_encode(self, X):
        if hasattr(self, 'ohc'):
            return self.ohc.transform(X)
        else:
            categorical_features = [c for c in X.columns if X[c].nunique() < 8]
            self.ohc = ColumnTransformer(
                [
                    (
                        "cat",
                        OneHotEncoder(
                            handle_unknown="ignore", 
                            sparse_output=False
                        ),
                        categorical_features,
                    ),
                ],
                verbose_feature_names_out=False,
                remainder='passthrough'
            )
            return self.ohc.fit_transform(X)

    def init(self, X, incremental=False):
        """Overload of MLP training. Just determines the dimensions of 
        the coefficients in the network. 
        Adapted from _multilayer_perceptron.py::_fit()

        """
        # one-hot encode X
        X = self._one_hot_encode(X)
        # make a random y
        y = np.random.randint(0,1,size=len(X))
        # Make sure self.hidden_layer_sizes is a list
        hidden_layer_sizes = self.hidden_layer_sizes
        if not hasattr(hidden_layer_sizes, "__iter__"):
            hidden_layer_sizes = [hidden_layer_sizes]
        hidden_layer_sizes = list(hidden_layer_sizes)

        if np.any(np.array(hidden_layer_sizes) <= 0):
            raise ValueError(
                "hidden_layer_sizes must be > 0, got %s." % hidden_layer_sizes
            )
        first_pass = not hasattr(self, "coefs_") or (
            not self.warm_start and not incremental
        )

        X, y = self._validate_input(X, y, incremental, reset=first_pass)

        n_samples, n_features = X.shape

        # Ensure y is 2D
        if y.ndim == 1:
            y = y.reshape((-1, 1))

        self.n_outputs_ = y.shape[1]

        layer_units = [n_features] + hidden_layer_sizes + [self.n_outputs_]

        # check random state
        self._random_state = check_random_state(self.random_state)

        if first_pass:
            # First time training the model
            self._initialize(y, layer_units, X.dtype)

        # Initialize lists
        activations = [X] + [None] * (len(layer_units) - 1)
        deltas = [None] * (len(activations) - 1)

        coef_grads = [
            np.empty((n_fan_in_, n_fan_out_), dtype=X.dtype)
            for n_fan_in_, n_fan_out_ in zip(layer_units[:-1], layer_units[1:])
        ]

        intercept_grads = [
            np.empty(n_fan_out_, dtype=X.dtype) for n_fan_out_ in layer_units[1:]
        ]

        # from _fit_lbfgs:
        # Store meta information for the parameters
        self._coef_indptr = []
        self._intercept_indptr = []
        start = 0

        # Save sizes and indices of coefficients for faster unpacking
        for i in range(self.n_layers_ - 1):
            n_fan_in, n_fan_out = layer_units[i], layer_units[i + 1]

            end = start + (n_fan_in * n_fan_out)
            self._coef_indptr.append((start, end, (n_fan_in, n_fan_out)))
            start = end

        # Save sizes and indices of intercepts for faster unpacking
        for i in range(self.n_layers_ - 1):
            end = start + layer_units[i + 1]
            self._intercept_indptr.append((start, end))
            start = end

        return self

class Linear:

    def __init__(self, Xp):
        # Xohc = pd.get_dummies(Xp.astype('category'))
        # print('Xohc shape:',Xohc.shape)
        X = self._one_hot_encode(Xp)
        self.coefs_ = np.empty(X.shape[1] + 1) 

    def get_n_weights(self):
        """Sets coefs_ and intercepts_ to x."""
        return len(self.coefs_)

    def set_weights(self, x):
        """Sets coefs_ and intercepts_ to x."""
        self.coefs_ = x

    def fit(self, X, y):
        raise NotImplementedError('Dont call fit on this class!')
        return self

    def predict(self, X):
        # Xohc = pd.get_dummies(X.astype('category'))
        X = self._one_hot_encode(X)
        intercept = np.ones(X.shape[0])
        Xintercept = np.column_stack((intercept, X))
        return expit(np.dot(Xintercept,self.coefs_))

    def _one_hot_encode(self, X):
        if hasattr(self, 'ohc'):
            return self.ohc.transform(X)
        else:
            binary_columns = [col for col in X.columns if X[col].isin([0, 1]).all()]
            categorical_features = [c for c in X.columns if (X[c].nunique() <= 8 and c not in binary_columns)] #Do not one-hot-encode binary columns and columns with more than 8 categories. 
            self.ohc = ColumnTransformer(
                [
                    (
                        "cat",
                        OneHotEncoder(
                            handle_unknown="ignore", 
                            sparse_output=False
                        ),
                        categorical_features,
                    ),
                ],
                verbose_feature_names_out=False,
                remainder='passthrough'
            )
            return self.ohc.fit_transform(X)

class InterLinear:
    """Make an interaction version of a linear model
    change one hot encode to capture all unique values of Xp
    """

    def __init__(self, Xp):
        # Xohc = pd.get_dummies(Xp.astype('category'))
        # print('Xohc shape:',Xohc.shape)
        # Xpu = Xp.unique()
        X = self._one_hot_encode(Xp)
        self.coefs_ = np.empty(X.shape[1] + 1) 

    def get_n_weights(self):
        """Sets coefs_ and intercepts_ to x."""
        return len(self.coefs_)

    def set_weights(self, x):
        """Sets coefs_ and intercepts_ to x."""
        self.coefs_ = x

    def fit(self, X, y):
        raise NotImplementedError('Dont call fit on this class!')
        return self

    def predict(self, X):
        # Xohc = pd.get_dummies(X.astype('category'))
        X = self._one_hot_encode(X)
        intercept = np.ones(X.shape[0])
        Xintercept = np.column_stack((intercept, X))
        return expit(np.dot(Xintercept,self.coefs_))

    def _make_interaction(self, X):
        """add an interaction variable"""
        X = X.copy()
        groups = list(X.columns)
        categories = X.groupby(groups).groups
        categories = {k:i for i,k in enumerate(categories.keys())}
        def categorize(x):
            return categories[tuple(x[c] for c in x.index)]
        X['interaction'] = X.apply(lambda x: categorize(x), axis=1)
        return X

    def _one_hot_encode(self, X):

        X = self._make_interaction(X)

        if not hasattr(self, 'ohc'):
            self.ohc = ColumnTransformer(
                [
                    (
                        "cat",
                        OneHotEncoder(
                            handle_unknown="ignore", 
                            sparse_output=False
                        ),
                        X.columns,
                    ),
                ],
                verbose_feature_names_out=False,
                remainder='passthrough'
            )
            self.ohc.fit(X)

        return self.ohc.transform(X)


"""
Fairness Oriented Multiobjective Optimization (Fomo)

BSD 3-Clause License

Copyright (c) 2023, William La Cava

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright notice, this
   list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright notice,
   this list of conditions and the following disclaimer in the documentation
   and/or other materials provided with the distribution.

3. Neither the name of the copyright holder nor the names of its
   contributors may be used to endorse or promote products derived from
   this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
"""
import numpy as np
from pymoo.core.problem import ElementwiseProblem
from sklearn.base import clone, ClassifierMixin
from sklearn.utils import resample
import warnings
import inspect
from .surrogate_models import MLP

class SurrogateProblem(ElementwiseProblem):
    """ The evaluation function for each candidate weights. 

    """

    def __init__( self, fomo_estimator, metric_kwargs={}, **kwargs):

        self.fomo_estimator=fomo_estimator
        self.metric_kwargs=metric_kwargs

        n_obj = (len(self.fomo_estimator.accuracy_metrics_)
                 +len(self.fomo_estimator.fairness_metrics_)
        )

        for k,v in self.metric_kwargs.items(): 
            print(k,v)
            if v is not None:
                if k == 'X_protected':
                    self.X_protected = v
                elif k == 'groups':
                    X = self.fomo_estimator.X_
                    self.X_protected = X[v]
                else:
                    print('no match for',k)
                break

        n_var = self._get_surrogate().get_n_weights()

        super().__init__(
            n_var = n_var,
            n_obj = n_obj,
            # set lower bound to -1
            xl = -np.ones(n_var),
            # set upper bound to 1
            xu = np.ones(n_var),
            **kwargs
        )

    def _get_surrogate(self):
        return MLP(hidden_layer_sizes=(10,)).init(self.X_protected)

    def get_sample_weight(self, x):
        surrogate = self._get_surrogate()
        surrogate.set_weights(x)
        sample_weight = surrogate.predict_proba(self.X_protected)[:,1]
        return sample_weight

    def _evaluate(self, x, out, *args, **kwargs):
        """Evaluate the weights by fitting an estimator and evaluating."""
        try:
            X = self.fomo_estimator.X_
            y = self.fomo_estimator.y_

            sample_weight = self.get_sample_weight(x)
            # sample_weight = sample_weight.ravel()
            assert len(sample_weight) == len(X)

            est = clone(self.fomo_estimator.estimator)
            with warnings.catch_warnings():
                warnings.simplefilter("ignore")
                est.fit(X,y,sample_weight=sample_weight)

            f = np.empty(self.n_obj)
            j = 0
            for i, metric in enumerate(self.fomo_estimator.accuracy_metrics_):
                f[i] = metric(est, X, y)
                j += 1
            for metric in self.fomo_estimator.fairness_metrics_:
                f[j] = metric(est, X, y, **self.metric_kwargs)
                j += 1

            out['F'] = np.asarray(f)
        except Exception as e:
            print(e)
            import ipdb
            ipdb.set_trace()


"""
Fairness Oriented Multiobjective Optimization (Fomo)

BSD 3-Clause License

Copyright (c) 2023, William La Cava

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright notice, this
   list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright notice,
   this list of conditions and the following disclaimer in the documentation
   and/or other materials provided with the distribution.

3. Neither the name of the copyright holder nor the names of its
   contributors may be used to endorse or promote products derived from
   this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
"""

import numpy as np
import pandas as pd

import random
#from pymoo.algorithms.base.genetic import GeneticAlgorithm
from pymoo.operators.selection.tournament import compare, TournamentSelection
from pymoo.algorithms.moo.nsga2 import NSGA2
from pymoo.algorithms.moo.nsga2 import binary_tournament
from fomo.utils import categorize
from pymoo.core.survival import Survival
from pymoo.core.selection import Selection
from pymoo.operators.sampling.rnd import FloatRandomSampling
from pymoo.operators.survival.rank_and_crowding import RankAndCrowding
from pymoo.util.display.multi import MultiObjectiveOutput
from pymoo.termination.default import DefaultMultiObjectiveTermination
from pymoo.util.misc import has_feasible
from pymoo.operators.crossover.sbx import SBX
from pymoo.operators.mutation.pm import PM
from pymoo.algorithms.base.genetic import GeneticAlgorithm

def get_parent(pop):

    if not hasattr(get_parent_WeightedCoinFlip, "_called"):
        print("Default flex")
        get_parent_WeightedCoinFlip._called = True
    
    fng = pop.get("fng")
    fn = pop.get("fn")
    G = np.arange(fng.shape[1])
    S = np.arange(len(pop))
    loss = []

    while (len(G) > 0 and len(S) > 1):

        g = random.choice(G)
        loss = []
        
        if (random.random() < 0.5):
            #half the time look at accuracy
            loss = fng[:, g]
        else:
            #half the time look at fairness
            loss = np.abs(fng[:, g] - fn)

        L = min(loss) 
        epsilon = np.median(np.abs(loss - np.median(loss)))
        survivors = np.where(loss <= L + epsilon)
        S = S[survivors]
        fng = fng[survivors] 
        fn = fn[survivors]
        G = G[np.where(G != g)]
            
    S = S[:, None].astype(int, copy=False)     
    return random.choice(S)
                
def get_parent_noCoinFlip(pop):

    if not hasattr(get_parent_WeightedCoinFlip, "_called"):
        print("Flex with no coin flip")
        get_parent_WeightedCoinFlip._called = True

    fng = pop.get("fng")
    fng = np.tile(fng, 2)
    fn = pop.get("fn")
    G = np.arange(fng.shape[1])
    S = np.arange(len(pop))
    loss = []

    while (len(G) > 0 and len(S) > 1):

        g = random.choice(G)
        loss = []
        
        if g < max(G)/2:
            #look at accuracy
            loss = fng[:, g]
        else:
            #look at fairness
            loss = np.abs(fng[:, g] - fn)

        L = min(loss) 
        epsilon = np.median(np.abs(loss - np.median(loss)))
        survivors = np.where(loss <= L + epsilon)
        S = S[survivors]
        fng = fng[survivors] 
        fn = fn[survivors]
        G = G[np.where(G != g)]

            
    S = S[:, None].astype(int, copy=False)     
    return random.choice(S)


def get_parent_WeightedCoinFlip(pop):

    if not hasattr(get_parent_WeightedCoinFlip, "_called"):
        print("Flex with weighted coin flip")
        get_parent_WeightedCoinFlip._called = True

    samples_fnr = pop.get("samples_fnr")
    fng = pop.get("fng")
    fn = pop.get("fn")
    gp_lens = pop.get('gp_lens')
    G = np.arange(fng.shape[1])
    S = np.arange(len(pop))
    loss = []
    weight = random.random()

    while (len(G) > 0 and len(S) > 1):

        g = random.choice(G)
        loss = []

        if (random.random() > weight):
            #look at fairness
            loss = fng[:, g]
            G = G[np.where(G != g)]
        else:
            #look at accuracy
            num_rows, num_cols = np.shape(samples_fnr)
            indices = np.random.choice(num_cols, size = int(gp_lens[0, g]), replace = False)
            fnr_sum = np.sum(samples_fnr[:, indices], axis=1)
            pos_count = np.sum(samples_fnr[:, indices].astype(bool), axis=1)
            for i in range (len(pos_count)):
                if pos_count[i]:
                    loss.append(fnr_sum[i]/pos_count[i])
                else:
                    loss.append(0)


        L = min(loss) 
        epsilon = np.median(np.abs(loss - np.median(loss)))
        survivors = np.where(loss <= L + epsilon)
        S = S[survivors]
        fng = fng[survivors] 
        fn = fn[survivors]
        samples_fnr = samples_fnr[survivors]
        gp_lens = gp_lens[survivors]
            
    S = S[:, None].astype(int, copy=False)     
    return random.choice(S)

class FLEX(Selection):
    
    def __init__(self,
                 **kwargs):

        #self.X_protected_ = X_protected
        #self.categories = categories
        #self.group_losses = group_losses
        super().__init__(**kwargs)
     
         
    def _do(self, _, pop, n_select, n_parents=1, flag = 0, **kwargs):

        # offss = super().n_offsprings
        # s = n_select * n_parents
        # S = np.random.randint(0, s, s)
        # S = S[:, None].astype(int, copy=False)
        
        parents = []
        
        for i in range(n_select * n_parents): 
            #get pop_size parents
            p = get_parent(pop)
            parents.append(p)
            
        return np.reshape(parents, (n_select, n_parents))

class LexSurvival(Survival):
    def __init__(self) -> None:
        super().__init__(filter_infeasible=False)

    def _do(self, problem, pop, n_survive=None, **kwargs):
        return pop[-n_survive:]


class Lexicase_NSGA2(GeneticAlgorithm):

    def __init__(self,
                 pop_size=100,
                 sampling=FloatRandomSampling(),
                 selection=FLEX(),
                 crossover=SBX(eta=15, prob=0.9),
                 mutation=PM(eta=20),
                 survival=RankAndCrowding(),
                 output=MultiObjectiveOutput(),
                 **kwargs):
        
        super().__init__(
            pop_size=pop_size,
            sampling=sampling,
            selection=selection,
            crossover=crossover,
            mutation=mutation,
            survival=survival,
            output=output,
            advance_after_initial_infill=True,
            **kwargs)
        
        self.termination = DefaultMultiObjectiveTermination()
        self.tournament_type = 'comp_by_dom_and_crowding'

    def _set_optimum(self, **kwargs):
        if not has_feasible(self.pop):
            self.opt = self.pop[[np.argmin(self.pop.get("CV"))]]
        else:
            self.opt = self.pop[self.pop.get("rank") == 0]

class Lexicase(GeneticAlgorithm):

    def __init__(self,
                 pop_size=100,
                 sampling=FloatRandomSampling(),
                 selection=FLEX(),
                 crossover=SBX(eta=15, prob=0.9),
                 mutation=PM(eta=20),
                 survival=LexSurvival(),
                 output=MultiObjectiveOutput(),
                 **kwargs):
        
        super().__init__(
            pop_size=pop_size,
            sampling=sampling,
            selection=selection,
            crossover=crossover,
            mutation=mutation,
            survival=survival,
            output=output,
            advance_after_initial_infill=True,
            **kwargs)


"""
Fairness Oriented Multiobjective Optimization (Fomo)

BSD 3-Clause License

Copyright (c) 2023, William La Cava

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright notice, this
   list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright notice,
   this list of conditions and the following disclaimer in the documentation
   and/or other materials provided with the distribution.

3. Neither the name of the copyright holder nor the names of its
   contributors may be used to endorse or promote products derived from
   this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
"""
import numpy as np
from pymoo.core.problem import ElementwiseProblem
from sklearn.base import clone, ClassifierMixin
from sklearn.utils import resample
from sklearn.pipeline import Pipeline
import warnings
import inspect
import fomo.metrics as metrics
from .surrogate_models import MLP, Linear, InterLinear
from fomo.algorithm import Lexicase, Lexicase_NSGA2

class BasicProblem(ElementwiseProblem):
    """ The evaluation function for each candidate sample weights. """

    def __init__(
        self, 
        fomo_estimator, 
        metric_kwargs={},
        **kwargs
        ):
        self.fomo_estimator=fomo_estimator
        self.metric_kwargs=metric_kwargs
        n_var = len(self.fomo_estimator.X_)
        n_obj = (len(self.fomo_estimator.accuracy_metrics_)
                  +len(self.fomo_estimator.fairness_metrics_)
        )
        
        super().__init__(
            n_var = n_var,
            n_obj = n_obj,
            xl = np.zeros(n_var),
            xu = np.ones(n_var),
            **kwargs
        )

    def get_sample_weight(self, x):
        return x

    def _evaluate(self, sample_weight, out, *args, **kwargs):
        """Evaluate the weights by fitting an estimator and evaluating."""

        X = self.fomo_estimator.X_
        y = self.fomo_estimator.y_
        est = clone(self.fomo_estimator.estimator)

        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            if isinstance(est, Pipeline):
                stepname = est.steps[-1][0]
                param_name = stepname + '__sample_weight'
                kwarg = {param_name:sample_weight}
                est.fit(X, y, **kwarg)
            else:
                est.fit(X,y,sample_weight=sample_weight)

        f = np.empty(self.n_obj)
        j = 0
        for i, metric in enumerate(self.fomo_estimator.accuracy_metrics_):
            f[i] = metric(est, X, y)
            j += 1
        for metric in self.fomo_estimator.fairness_metrics_:
            f[j] = metric(est, X, y, **self.metric_kwargs)
            j += 1
            
        out['F'] = np.asarray(f)

        if isinstance(self.fomo_estimator.algorithm, (Lexicase, Lexicase_NSGA2)):
            fn, fng, samples_fnr, gp_lens = metrics.flex_loss(est, X, y, 'FNR', **self.metric_kwargs)
            out['fn'] = fn #FNR of all samples to be used in Flex
            out['fng'] = fng #FNR of every group to be used in Flex
            out['samples_fnr'] = samples_fnr #FNR of each sample to be used in Flex with weighted coin flip
            out['gp_lens'] = gp_lens #Length of each protected group to be used in Flex with weighted coin flip


class SurrogateProblem(ElementwiseProblem):
    """ The evaluation function for each candidate weights. 

    """

    def __init__( self, fomo_estimator, metric_kwargs={}, **kwargs):

        self.fomo_estimator=fomo_estimator
        self.metric_kwargs=metric_kwargs

        n_obj = (len(self.fomo_estimator.accuracy_metrics_)
                 +len(self.fomo_estimator.fairness_metrics_)
        )

        for k,v in self.metric_kwargs.items(): 
            print(k,v)
            if v is not None:
                if k == 'X_protected':
                    self.X_protected = v
                elif k == 'groups':
                    X = self.fomo_estimator.X_
                    self.X_protected = X[v]
                else:
                    print('no match for',k)
                break

        n_var = self._get_surrogate().get_n_weights()

        super().__init__(
            n_var = n_var,
            n_obj = n_obj,
            # set lower bound to -1
            xl = -np.ones(n_var),
            # set upper bound to 1
            xu = np.ones(n_var),
            **kwargs
        )

    def _get_surrogate(self):
        raise NotImplementedError('Dont call _get_surrogate on this base class!')

    def get_sample_weight(self, x):
        surrogate = self._get_surrogate()
        surrogate.set_weights(x)
        sample_weight = surrogate.predict(self.X_protected)
        return sample_weight

    def _evaluate(self, x, out, *args, **kwargs):
        """Evaluate the weights by fitting an estimator and evaluating."""
        X = self.fomo_estimator.X_
        y = self.fomo_estimator.y_

        sample_weight = self.get_sample_weight(x)
        # sample_weight = sample_weight.ravel()
        assert len(sample_weight) == len(X)

        est = clone(self.fomo_estimator.estimator)
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            if isinstance(est, Pipeline):
                stepname = est.steps[-1][0]
                param_name = stepname + '__sample_weight'
                kwarg = {param_name:sample_weight}
                est.fit(X, y, **kwarg)
            else:
                est.fit(X,y,sample_weight=sample_weight)

        f = np.empty(self.n_obj)
        j = 0
        for i, metric in enumerate(self.fomo_estimator.accuracy_metrics_):
            f[i] = metric(est, X, y)
            j += 1
        for metric in self.fomo_estimator.fairness_metrics_:
            f[j] = metric(est, X, y, **self.metric_kwargs)
            j += 1

        out['F'] = np.asarray(f)

        if isinstance(self.fomo_estimator.algorithm, (Lexicase, Lexicase_NSGA2)):
            fn, fng, samples_fnr, gp_lens = metrics.flex_loss(est, X, y, 'FNR', **self.metric_kwargs)
            out['fn'] = fn #FNR of all samples to be used in Flex
            out['fng'] = fng #FNR of every group to be used in Flex
            out['samples_fnr'] = samples_fnr #FNR of each sample to be used in Flex with weighted coin flip
            out['gp_lens'] = gp_lens #Length of each protected group to be used in Flex with weighted coin flip

class MLPProblem(SurrogateProblem):
    """ The evaluation function for each candidate weights. 

    """
    def _get_surrogate(self):
        return MLP(hidden_layer_sizes=(10,)).init(self.X_protected)

class LinearProblem(SurrogateProblem):
    """ The evaluation function for each candidate weights. 

    """
    def _get_surrogate(self):
        return Linear(self.X_protected)

class InterLinearProblem(SurrogateProblem):
    """ The evaluation function for each candidate weights. 

    """
    def _get_surrogate(self):
        return InterLinear(self.X_protected)



"""
Fairness Oriented Multiobjective Optimization (Fomo)

BSD 3-Clause License

Copyright (c) 2023, William La Cava

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright notice, this
   list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright notice,
   this list of conditions and the following disclaimer in the documentation
   and/or other materials provided with the distribution.

3. Neither the name of the copyright holder nor the names of its
   contributors may be used to endorse or promote products derived from
   this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
"""
import numpy as np
import pandas as pd

def squash_array(x):
    x[x<0.0] == 0.0
    x[x>1.0] == 1.0
    return x

def squash_series(x):
    return x.apply(lambda x: max(x,0.0)).apply(lambda x: min(x,1.0))

def category_diff(cat1, cat2):
    different=False
    for k1,v1 in cat1.items():
        if k1 not in cat2.keys():
            print(f'{k1} not in cat2')
            different=True
        else:
            if not v1.equals(cat2[k1]):
                print(f'indices for {k1} different in cat2')
                different=True
    for k2,v2 in cat2.items():
        if k2 not in cat1.keys():
            print(f'{k1} not in cat1')
            different=True
        else:
            if not v2.equals(cat1[k2]):
                print(f'indices for {k2} different in cat1')
                different=True
    if not different:
        # print('categories match.')
        return True
    else:
        return False

def categorize(X, y, groups, grouping,
               n_bins=10,
               bins=None,
               alpha=0.01,
               gamma=0.01
              ):
    """Map data to an existing set of categories."""
    assert isinstance(X, pd.DataFrame), "X should be a dataframe"

    categories = None 

    if bins is None:
        if n_bins is None:
            n_bins = 10
        bins = np.linspace(float(1.0/n_bins), 1.0, n_bins)
        bins[0] = 0.0
    else:
        n_bins=len(bins)


    df = X[groups].copy()
    df.loc[:,'interval'], retbins = pd.cut(y, bins, 
                                           include_lowest=True,
                                           retbins=True
                                          )
    categories = {}
    # group_ids = groupby(X, y, groups, grouping)
    group_ids = df.groupby(groups).groups

    min_grp_size = gamma*len(X) 
    min_cat_size = min_grp_size*alpha/n_bins
    for group, i in group_ids.items():
        # filter groups smaller than gamma*len(X)
        if len(i) <= min_grp_size:
            continue
        for interval, j in df.loc[i].groupby('interval').groups.items():
            # filter categories smaller than alpha*gamma*len(X)/n_bins
            if len(j) > min_cat_size:
                categories[group + (interval,)] = j
    return categories

from pymoo.decomposition.asf import ASF
class Compromise:
    def __init__(self, weights):
        self.weights=weights
    def do(self,F):
        decomp = ASF()
        I = decomp(F, self.weights).argmin()
        return I


"""
Fairness Oriented Multiobjective Optimization (Fomo)

BSD 3-Clause License

Copyright (c) 2023, William La Cava

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright notice, this
   list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright notice,
   this list of conditions and the following disclaimer in the documentation
   and/or other materials provided with the distribution.

3. Neither the name of the copyright holder nor the names of its
   contributors may be used to endorse or promote products derived from
   this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
"""
import warnings
import ipdb
import numpy as np
import pandas as pd
import logging
import itertools as it
from fomo.utils import categorize 
from sklearn.metrics import mean_squared_error

logger = logging.getLogger(__name__)

def pairwise(iterable):
    "s -> (s0,s1), (s1,s2), (s2, s3), ..."
    a, b = it.tee(iterable)
    next(b, None)
    return zip(a, b)

def stratify_groups(X, y, groups,
               n_bins=10,
               bins=None,
               alpha=0.0,
               gamma=0.0
              ):
    """Map data to an existing set of groups, stratified by risk interval."""
    assert isinstance(X, pd.DataFrame), "X should be a dataframe"


    if bins is None:
        bins = np.linspace(float(1.0/n_bins), 1.0, n_bins)
        bins[0] = 0.0
    else:
        n_bins=len(bins)


    df = X[groups].copy()
    df.loc[:,'interval'], retbins = pd.cut(y, bins, 
                                           include_lowest=True,
                                           retbins=True
                                          )
    stratified_categories = {}
    min_size = gamma*alpha*len(X)/n_bins
    for group, dfg in df.groupby(groups):
        # filter groups smaller than gamma*len(X)
        if len(dfg)/len(X) <= gamma:
            continue
        
        for interval, j in dfg.groupby('interval').groups.items():
            if len(j) > min_size:
                if interval not in stratified_categories.keys():
                    stratified_categories[interval] = {}

                stratified_categories[interval][group] = j
    # now we have categories where, for each interval, there is a dict of groups.
    return stratified_categories

def multicalibration_loss(
    estimator,
    X,
    y_true,
    groups=None,
    X_protected=None,
    grouping='intersectional',
    n_bins=None,
    bins=None,
    categories=None,
    proportional=False,
    alpha=0.01,
    gamma=0.01,
    rho=0.1
):
    """custom scoring function for multicalibration.
       calculate current loss in terms of (proportional) multicalibration
    """
    if not isinstance(y_true, pd.Series):
        y_true = pd.Series(y_true)

    y_pred = estimator.predict_proba(X)[:,1]
    y_pred = pd.Series(y_pred, index=y_true.index)


    assert isinstance(y_true, pd.Series)
    assert isinstance(y_pred, pd.Series)
    loss = 0.0

    assert groups is not None or X_protected is not None, "groups or X_protected must be defined."

    if categories is None:
        categories = categorize(X, y_pred, groups, grouping,
                                n_bins=n_bins,
                                bins=bins,
                                alpha=alpha, 
                                gamma=gamma
                               )

    for c, idx in categories.items():
        category_loss = np.abs(y_true.loc[idx].mean() 
                               - y_pred.loc[idx].mean()
                              )
        if proportional: 
            category_loss /= max(y_true.loc[idx].mean(), rho)

        if  category_loss > loss:
            loss = category_loss

    return loss

def multicalibration_score(estimator, X, y_true, **kwargs):
    return -multicalibration_loss(estimator, X, y_true, **kwargs)

def proportional_multicalibration_loss(estimator, X, y_true, **kwargs):
    kwargs['proportional'] = True
    return multicalibration_loss(estimator, X, y_true, **kwargs)

def proportional_multicalibration_score(estimator, X, y_true, groups, **kwargs):
    return -proportional_multicalibration_loss(estimator, X, y_true, groups,  **kwargs)

def differential_calibration_loss(
    estimator, 
    X, 
    y_true,
    groups=None,
    X_protected=None,
    n_bins=None,
    bins=None,
    stratified_categories=None,
    alpha=0.0,
    gamma=0.0,
    rho=0.0
):
    """Return the differential calibration of estimator on groups."""

    assert groups is not None or X_protected is not None, "groups or X_protected must be defined."
    assert isinstance(X, pd.DataFrame), "X needs to be a dataframe"
    assert all([g in X.columns for g in groups]), ("groups not found in"
                                                   " X.columns")
    if not isinstance(y_true, pd.Series):
        y_true = pd.Series(y_true)

    y_pred = estimator.predict_proba(X)[:,1]

    if stratified_categories is None:
        stratified_categories = stratify_groups(X, y_pred, groups,
                                n_bins=n_bins,
                                bins=bins,
                                alpha=alpha, 
                                gamma=gamma
                               )
    logger.info(f'# categories: {len(stratified_categories)}')
    dc_max = 0
    logger.info("calclating pairwise differential calibration...")
    for interval in stratified_categories.keys():
        for (ci,i),(cj,j) in pairwise(stratified_categories[interval].items()):

            yi = max(y_true.loc[i].mean(), rho)
            yj = max(y_true.loc[j].mean(), rho)

            dc = np.abs( np.log(yi) - np.log(yj) )

            if dc > dc_max:
                dc_max = dc

    return dc_max

def differential_calibration_score(estimator, X, y_true, **kwargs):
    return -differential_calibration_loss(estimator, X, y_true, **kwargs)

def TPR(y_true, y_pred):
    """Returns True Positive Rate.

    Parameters
    ----------
    y_true: array-like, bool 
        True labels. 
    y_pred: array-like, float or bool
        Predicted labels. 

    If y_pred is floats, this is the "soft" true positive rate 
    (i.e. the average probability estimate for the positive class)
    """
    return np.sum(y_pred[(y_true==1)])/np.sum(y_true)

def FPR(y_true, y_pred):
    """Returns False Positive Rate.

    Parameters
    ----------
    y_true: array-like, bool 
        True labels. 
    y_pred: array-like, float or bool
        Predicted labels. 

    If y_pred is floats, this is the "soft" false positive rate 
    (i.e. the average probability estimate for the negative class)
    """
    # if there are no negative labels, return zero
    if np.sum(y_true) == len(y_true):
        return 0
    yt = y_true.astype(bool)
    return np.sum(y_pred[~yt])/np.sum(~yt)

def FNR(y_true, y_pred):
    """Returns False Negative Rate.

    Parameters
    ----------
    y_true: array-like, bool 
        True labels. 
    y_pred: array-like, float or bool
        Predicted labels. 

    If y_pred is floats, this is the "soft" false negative rate 
    (i.e. the average probability estimate for the negative class)
    """
    # if there are no postive labels, return zero
    if np.sum(y_true) == 0:
        return 0
    yt = y_true.astype(bool)
    return np.sum(1-y_pred[yt])/np.sum(yt)


def subgroup_loss(y_true, y_pred, X_protected, metric, grouping = 'intersectional', abs_val = False, gamma = True):
    assert isinstance(X_protected, pd.DataFrame), "X should be a dataframe"
    if not isinstance(y_true, pd.Series):
        y_true = pd.Series(y_true, index=X_protected.index)
    else:
        y_true.index = X_protected.index

    y_pred = pd.Series(y_pred, index=X_protected.index)

    if (grouping == 'intersectional'):
        groups = list(X_protected.columns)
        categories = X_protected.groupby(groups).groups  
    else:
        categories = {}
        for col in X_protected.columns:
            unique_values = X_protected[col].unique()
            for val in unique_values:
                category_key = f'{col}_{val}'
                mask = X_protected[col] == val
                indices = X_protected[mask].index
                categories[category_key] = indices
    # print('#intersectional groups: ', len(categories))
    # singles = 0
    # gp_lens = [len(lst) for lst in categories.values()]
    # singles = gp_lens.count(1)
    # avg_len = sum(gp_lens) / len(gp_lens) if gp_lens else 0

    if isinstance(metric,str):
        loss_fn = FPR if metric=='FPR' else FNR
    elif callable(metric):
        loss_fn = metric
    else:
        raise ValueError(f'metric={metric} must be "FPR", "FNR", or a callable')

    base_loss = loss_fn(y_true, y_pred)
    max_loss = 0.0
    for c, idx in categories.items():
        # for FPR and FNR, gamma is also conditioned on the outcome probability
        if metric=='FPR' or loss_fn == FPR: 
            g = 1 - np.sum(y_true.loc[idx])/len(X_protected)
        elif metric=='FNR' or loss_fn == FNR: 
            g = np.sum(y_true.loc[idx])/len(X_protected)
        else:
            g = len(idx) / len(X_protected)

        category_loss = loss_fn(
            y_true.loc[idx].values, 
            y_pred.loc[idx].values
        )
        
        deviation = category_loss - base_loss

        if abs_val:
            deviation = np.abs(deviation)
        
        if gamma:
            deviation *= g

        if deviation > max_loss:
            max_loss = deviation

    return max_loss

def subgroup_FPR_loss(y_true, y_pred, X_protected, grouping = 'intersectional', abs_val = False, gamma = True):
    return subgroup_loss(y_true, y_pred, X_protected, 'FPR', grouping, abs_val, gamma)

def subgroup_FNR_loss(y_true, y_pred, X_protected, grouping = 'intersectional', abs_val = False, gamma = True):
    return subgroup_loss(y_true, y_pred, X_protected, 'FNR', grouping, abs_val, gamma)

def subgroup_MSE_loss(y_true, y_pred, X_protected, grouping = 'intersectional', abs_val = False, gamma = True):
    return subgroup_loss(y_true, y_pred, X_protected, mean_squared_error, grouping, abs_val, gamma)

def subgroup_scorer(
    estimator,
    X,
    y_true,
    metric,
    grouping,
    abs_val,
    gamma,
    groups=None,
    X_protected=None,
    weights=None, 
):
    """Calculate the subgroup fairness of estimator on X according to `metric'.
    TODO: handle use case when Xp is passed
    """
    assert isinstance(X, pd.DataFrame), "X should be a dataframe"
    assert groups is not None or X_protected is not None, "groups or X_protected must be defined."

    y_pred = estimator.predict_proba(X)[:,1]
    # y_pred = estimator.predict(X)

    # assert groups is not None, "groups must be defined."
    if groups is None:
        assert X_protected is not None, "cannot define both groups and X_protected"
    else:
        assert X_protected is None, "cannot define both groups and X_protected"
        X_protected = X[groups]

    return subgroup_loss(y_true, y_pred, X_protected, metric, grouping, abs_val, gamma)

def subgroup_FPR_scorer(estimator, X, y_true, **kwargs):
    return subgroup_scorer( estimator, X, y_true, 'FPR', **kwargs)

def subgroup_FNR_scorer(estimator, X, y_true, **kwargs):
    return subgroup_scorer( estimator, X, y_true, 'FNR', **kwargs)

def subgroup_MSE_scorer(estimator, X, y_true, **kwargs):
    return subgroup_scorer( estimator, X, y_true, mean_squared_error, **kwargs)


def flex_loss(estimator, X, y_true, metric, **kwargs):
    """
        returns 
        ----------
        fn: overall loss of all samples
        fng: loss over group for every group in the training data
        samples_fnr: False negative rate of every sample in the training data
        gp_lens: length of each protected group
        
        Parameters
        ----------
        estimator : sklearn-like estimator
            The underlying ML model to be trained.
        X : array-like, shape (n_samples, n_features)
            The training input samples.
        y_true: array-like, bool 
            True labels.
        metric: string or function
            The loss function. Could be FPR or FNR.
        flag: bool
            flag = 1 means marginal grouping and flag = 0 means intersectional grouping
    """
    
    groups = kwargs['groups']
    X_protected = X[groups]
    categories = {}
    fng = []
    samples_fnr = []
    gp_lens = []
    
    y_pred = estimator.predict_proba(X)[:,1]
    y_pred = pd.Series(y_pred, index=X_protected.index)

    if isinstance(metric,str):
        loss_fn = FPR if metric=='FPR' else FNR
    elif callable(metric):
        loss_fn = metric
    else:
        raise ValueError(f'metric={metric} must be "FPR", "FNR", or a callable')

    categories = {}
    for col in X_protected.columns:
        unique_values = X_protected[col].unique()
        for val in unique_values:
            category_key = f'{col}_{val}'
            mask = X_protected[col] == val
            indices = X_protected[mask].index
            categories[category_key] = indices
         
    for c, idx in categories.items():

        category_loss = loss_fn(
            y_true.loc[idx].values, 
            y_pred.loc[idx].values
        )
        fng.append(category_loss)
        gp_lens.append(len(y_true.loc[idx].values))

    # print('#marginal groups: ', len(categories))
    # singles = 0
    # singles = gp_lens.count(1)
    # avg_len = sum(gp_lens) / len(gp_lens) if gp_lens else 0

    #Calculate FNR of each sample
    for idx in y_true.index:
        fnr = loss_fn(y_true[idx], y_pred[idx])
        samples_fnr.append(fnr)

    fn = loss_fn(y_true, y_pred)    
    return fn, fng, samples_fnr, gp_lens


def mce(estimator, X, y_true, num_bins=10):
    """
        The metric to use if fairness is calibration-based.
        Returns maximum calibration error among the bins. 
        
        Parameters
        ----------
        estimator : sklearn-like estimator
            The underlying ML model to be trained.
        X : array-like, shape (n_samples, n_features)
            The training input samples.
        y_true: array-like, bool 
            True labels.
        num_bins: int
            Number of bins that the predictions are sorted and partitioned into. 
    """
    y_pred = estimator.predict_proba(X)[:,1]

    mce = 0
    for i in range(1, num_bins + 1):

        calibration_error = np.abs(y_true.mean() - y_pred.mean())
        mce = max(mce, calibration_error)

    return mce



"""
Fairness Oriented Multiobjective Optimization (Fomo)

BSD 3-Clause License

Copyright (c) 2023, William La Cava

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright notice, this
   list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright notice,
   this list of conditions and the following disclaimer in the documentation
   and/or other materials provided with the distribution.

3. Neither the name of the copyright holder nor the names of its
   contributors may be used to endorse or promote products derived from
   this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
"""
import copy
import math
import uuid 
import numpy as np
import pandas as pd
from sklearn.base import BaseEstimator, ClassifierMixin, RegressorMixin
from sklearn.utils.validation import check_is_fitted
from sklearn.utils.multiclass import unique_labels
from sklearn.metrics import make_scorer, roc_auc_score, mean_squared_error
from sklearn.linear_model import LogisticRegression, SGDRegressor
from sklearn.base import clone
from sklearn.pipeline import Pipeline
import multiprocessing
import dill

# pymoo
from pymoo.core.algorithm import Algorithm
from pymoo.core.problem import StarmapParallelization, ElementwiseProblem
from pymoo.algorithms.moo.nsga2 import NSGA2
from pymoo.optimize import minimize
# MCDM imports
from pymoo.mcdm.high_tradeoff import HighTradeoffPoints
from pymoo.mcdm.pseudo_weights import PseudoWeights
# fomo
from fomo.utils import Compromise
import fomo.metrics as metrics
from .problem import BasicProblem
# plotting
from pymoo.visualization.scatter import Scatter
# types
from collections.abc import Callable


class FomoEstimator(BaseEstimator):
    """ The base estimator for training fair models.  
    This class should not be called directly. Use :class:`FomoRegressor` or
    :class:`FomoClassifier` instead. 

    Parameters
    ----------
    estimator : sklearn-like estimator
        The underlying ML model to be trained. 
        The ML model must accept `sample_weight` as an argument to :meth:`fit`. 
    fairness_metrics : list[Callable]
        The fairness metrics to try to optimize during fitting. 
    accuracy_metrics : list[Callable]
        The accuracy metrics to try to optimize during fitting. 
    algorithm: pymoo Algorithm
        The multi-objective optimizer to use. Should be compatible with 
        `pymoo.core.algorithm.Algorithm`. 
    random_state: int | None
        Random seed. 
    verbose: bool
        Whether to print progress.
    n_jobs: int
        Number of parallel processes to use. Parallelizes evaluation.
    store_final_models: bool
        If True, the final set of models will be stored in the estimator.
    problem_type: ElementwiseProblem
        Determines the evaluation class to be used. Options:
        - :class:`BasicProblem`
        - :class:`MLPProblem`
        - :class:`LinearProblem`
    """
    def __init__(self, 
                 estimator: BaseEstimator,
                 fairness_metrics: list[Callable],
                 accuracy_metrics: list[Callable],
                 algorithm: Algorithm,
                 random_state: int,
                 verbose:bool,
                 n_jobs:int,
                 store_final_models:bool,
                 problem_type:ElementwiseProblem,
                 checkpoint:bool,
                 picking_strategy: str = 'PseudoWeights'
                ):
         self.estimator=estimator
         self.fairness_metrics=fairness_metrics
         self.accuracy_metrics=accuracy_metrics
         self.algorithm=algorithm
         self.random_state=random_state
         self.verbose=verbose
         self.n_jobs=n_jobs
         self.store_final_models=store_final_models
         self.problem_type=problem_type
         self.checkpoint=checkpoint
         self.picking_strategy=picking_strategy

    def fit(self, 
            X, 
            y, 
            grouping = 'intersectional', 
            abs_val = False, 
            gamma = True, 
            protected_features=None, 
            Xp=None, 
            starting_point=None, 
            **kwargs
        ):
        """Train the model.


        Parameters
        ----------
        X : array-like, shape (n_samples, n_features)
            The training input samples.
        y : array-like, shape (n_samples,) or (n_samples, n_outputs)
            The target values (class labels in classification, real numbers in
            regression).
        protected_columns: list[str] | None
            The columns in DataFrame X used to assign fairness. 
        Xp : {array-like, sparse matrix}, shape (n_samples, n_protected_features)
            The input samples for measuring/optimizing fairness during training.
        starting_point : str | None
            Optionally start from a checkpoint file with this name
        **kwargs : keyword arguments that are passed to `pymoo.optimize.minimize`.

        Returns
        -------
        self : object
            Returns self.
        """
        self._init_model()
        self.n_obj_ = len(self.accuracy_metrics_)+len(self.fairness_metrics_)
        ########################################
        # define problem
        # metric arguments
        metric_kwargs = dict(
            groups=protected_features, 
            X_protected=Xp,
            grouping = grouping, 
            abs_val = abs_val,
            gamma = gamma
        )
        problem_kwargs=dict(fomo_estimator=self, metric_kwargs=metric_kwargs)
        # parallelization
        n_processes = self.n_jobs if self.n_jobs > 0 else multiprocessing.cpu_count()
        print('running',n_processes,'processes')
        if n_processes > 1:
            pool = multiprocessing.Pool(n_processes)
            runner = StarmapParallelization(pool.starmap)
            problem_kwargs.update(dict(elementwise_runner=runner))

        self.problem_ = self.problem_type(**problem_kwargs)
        print('number of variables:',self.problem_.n_var)
        print('number of objectives:',self.problem_.n_obj)

        # define algorithm
        if starting_point is not None:
            with open(starting_point, 'rb') as f:
                self.algorithm_ = dill.load(f)
                print("Loaded Checkpoint:", self.algorithm_)
        else:
            self.algorithm_ = self.algorithm
        ########################################
        # minimize
        if self.checkpoint:
            run_id = uuid.uuid4()
            checkpoint_file = f"checkpoint.{run_id}.pkl"
            print('checkpoint file:',checkpoint_file)
            self.algorithm_.setup(
                self.problem_,
                seed=self.random_state,
                verbose=self.verbose,
                **kwargs
            )
            while self.algorithm_.has_next():
                self.algorithm_.next()
                with open(checkpoint_file, "wb") as f:
                    dill.dump(self.algorithm_, f)
            self.res_ = self.algorithm_.result()
        else:
            self.res_ = minimize(
                self.problem_,
                self.algorithm_,
                seed=self.random_state,
                verbose=self.verbose,
                **kwargs
            )

        if n_processes > 1:
            pool.close()
        ########################################
        # choose "best" estimator
        self.best_estimator_ = self.pick_best(strategy=self.picking_strategy) 
        self.is_fitted_ = True
        # store archive of estimators
        if self.store_final_models:
            self.estimator_archive_ = self._store_final_models()
        return self

    def _store_final_models(self):
        """Store archive of fitted estimators using final weights."""
        estimator_archive_ = []
        for x in self.res_.X:
            sample_weight = self.problem_.get_sample_weight(x)
            est = clone(self.estimator)
            if isinstance(est, Pipeline):
                stepname = est.steps[-1][0]
                param_name = stepname + '__sample_weight'
                kwarg = {param_name:sample_weight}
                est.fit(self.X_, self.y_, **kwarg)
            else:
                est.fit(self.X_, self.y_, sample_weight=sample_weight)
            estimator_archive_.append(est)
        return estimator_archive_

    def _output_archive(self, fn, **kwargs):
        """Call a function on every estimator in the archive and return output."""
        return [getattr(est, fn)(**kwargs) for est in self.estimator_archive_]

    def predict_archive(self, X):
        """Return a list of predictions from the archive models. """
        check_is_fitted(self, 'is_fitted_')
        if not hasattr(self, 'estimator_archive_'):
            print("Need to fit archive models. You can set `store_final_models` to True to avoid this step.")
            self.estimator_archive_ = self._store_final_models()

        return self._output_archive('predict', X=X)

        
    def pick_best(self, strategy='PseudoWeights', weights=None):
        """Picks the best solution based on on a multi-criteria decision-making
        (MCDM) strategy. 

        A description of MCDM strategies is given in the `pymoo docs <https://pymoo.org/mcdm/index.html>`_. 

        Parameters
        ----------
        strategy : str
            Name of an MCDM strategy. Built-in support for the following:

            - 'HighTradeOffPoints' : return a point near a cutoff.
            - 'Compromise': equally weight both objectives.  
            - 'PseudoWeights' : normalized weighting of both objectives. 

        weights: np.ndarray|None
            Weights for each objective. Used for Compromise and PseudoWeights methods.
            Default is equal weighting.
        """

        if type(weights) == type(None):
            if strategy in ['PseudoWeights','Compromise']:
                weights = np.array([float(1.0/self.n_obj_) for n in range(self.n_obj_)])

        if strategy == 'PseudoWeights':
            picking_fn = PseudoWeights(weights).do
        elif strategy == 'Compromise':
            picking_fn = Compromise(weights).do
        else:
            picking_fn = HighTradeoffPoints()

        F = self.res_.F.copy() 

        if len(F) <= 1:
            print('Warning: only one point on pareto front')
            I = 0
        else:
            I = picking_fn(F)
            if isinstance(I, np.ndarray):
                if len(I) > 1:
                    I = I[math.floor(len(I)/2)]
                else:
                    I = I[0]
            elif I is None:
                print('warning: picking returned None')
                I = np.random.randint(len(F))
        self.best_weights_ = self.res_.X[I]
        self.I_ = I
        best_est = clone(self.estimator)
        sample_weight = self.problem_.get_sample_weight(self.best_weights_)

        if isinstance(best_est, Pipeline):
            stepname = best_est.steps[-1][0]
            param_name = stepname + '__sample_weight'
            kwarg = {param_name:sample_weight}
            best_est.fit(self.X_, self.y_, **kwarg)
        else:
            best_est.fit(self.X_, self.y_, sample_weight=sample_weight)
        return best_est

    def predict(self, X):
        """Predict from X.

        Parameters
        ----------
        X : {array-like, sparse matrix}, shape (n_samples, n_features)
            The training input samples.

        Returns
        -------
        y : ndarray, shape (n_samples,)
        """
        # X = check_array(X, accept_sparse=True)
        check_is_fitted(self, 'is_fitted_')
        return self.best_estimator_.predict(X)

    def plot(self):
        """Plots the Pareto set of models with a dot indicating the selected 
        model.
        Returns
        -------
        plot : matplotlib figure
            The figure object.
        """
        check_is_fitted(self, 'is_fitted_')
        I = self.I_
        F = self._get_signed_F()
        axis_labels = self._get_objective_names()
        plot = (
            Scatter()
            .add(F, alpha=0.2, label='Candidate models')
            .add(F[I], color="red", s=100, label='Chosen model')
        )
        plot.axis_labels = axis_labels
        return plot

    def _init_model(self):
        if hasattr(self.estimator, 'random_state'):
            self.estimator.random_state = self.random_state
        if hasattr(self.estimator, 'n_jobs'):
            self.estimator.n_jobs = 1
    
    def _get_signed_F(self, F=None):
        if F is None:
            F = copy.copy(self.res_.F)
        # reverse F for metrics where higher is better
        for i,m in enumerate(self.accuracy_metrics_ + self.fairness_metrics_): 
            if hasattr(m, '_sign'):
                F[:,i] = F[:,i]*m._sign
        return F

    def _get_objective_names(self):
        """Returns names of functions defining the objectives"""
        labels = (
            [ m._score_func.__name__ for m in self.accuracy_metrics_ ] 
            + [ fn.__name__ for fn in self.fairness_metrics_ ]
        )
        labels = [l.replace('_',' ') for l in labels]
        return labels

    def get_pareto_points(self):
        """Return a Pandas dataframe of the Pareto archive points"""
        F = self._get_signed_F() 
        I = self.I_
        archive = pd.DataFrame(F, columns=self._get_objective_names())
        archive['chosen'] = [all(f==F[I]) for f in F]
        return archive

class FomoClassifier(FomoEstimator, ClassifierMixin, BaseEstimator):
    """FOMO Classifier. 

        1. Train a population of self.estimator models with random weights. 
        2. Update sample weights using self.algorithm. 
        3. Select a given model as best, but also save the set of models. 

    Parameters
    ----------
    estimator : sklearn-like estimator
        The underlying ML model to be trained. 
        The ML model must accept `sample_weight` as an argument to :meth:`fit`. 
    fairness_metrics : list[Callable]
        The fairness metrics to try to optimize during fitting. 
    accuracy_metrics : list[Callable]
        The accuracy metrics to try to optimize during fitting. 
    algorithm: pymoo Algorithm
        The multi-objective optimizer to use. Should be compatible with 
        `pymoo.core.algorithm.Algorithm`. 
    random_state: int | None
        Random seed. 
    verbose: bool
        Whether to print progress.
    n_jobs: int
        Number of parallel processes to use. Parallelizes evaluation.
    store_final_models: bool
        If True, the final set of models will be stored in the estimator.
    problem_type: ElementwiseProblem
        Determines the evaluation class to be used. Options:
        - :class:`BasicProblem`
        - :class:`MLPProblem`
        - :class:`LinearProblem`

    Examples
    --------
    >>> from fomo import FomoClassifier
    >>> from pmlb import pmlb
    >>> X,y = pmlb.fetch_data('adult', return_X_y=True)
    >>> groups = ['race','sex']
    >>> est = FomoClassifier()
    >>> est.fit(X,y, protected_features=groups)
    """
    def __init__(self, 
                 estimator: ClassifierMixin=LogisticRegression(),
                 fairness_metrics=None,
                 accuracy_metrics=None,
                 algorithm: Algorithm = NSGA2(),
                 random_state: int=None,
                 verbose: bool = False,
                 n_jobs: int = -1,
                 store_final_models: bool = False,
                 problem_type = BasicProblem,
                 checkpoint = False,
                 picking_strategy='PseudoWeights'
                ):
        super().__init__(
            estimator, 
            fairness_metrics, 
            accuracy_metrics,
            algorithm, 
            random_state,
            verbose,
            n_jobs,
            store_final_models,
            problem_type,
            checkpoint,
            picking_strategy
        )

    def fit(self, X, y, grouping = 'intersectional', abs_val = False, gamma = True, protected_features=None, Xp=None, **kwargs):
        """Train the model.

        1. Train a population of self.estimator models with random weights. 
        2. Update sample weights using self.algorithm. 
        3. Select a given model as best, but also save the set of models. 

        Parameters
        ----------
        X : array-like, shape (n_samples, n_features)
            The training input samples.
        y : array-like, shape (n_samples,)
            The target values. An array of int.
        protected_features: list[str]|None, default = None
            The columns of X to calculate fairness over. If specifying columns,
            do not also specify `Xp`.
        Xp: pandas DataFrame, shape (n_samples, n_protected_features), default=None
            The protected feature values used to calculate fairness. If `Xp` is 
            specified, `protected_features` must be None. 
        **kwargs : passed to `pymo.optimize.minimize`. 

        Returns
        -------
        self : object
            Returns self.
        """
        # Check that X and y have correct shape
        # X, y = check_X_y(X, y)
        # Store the classes seen during fit
        self.classes_ = unique_labels(y)

        self.X_ = X
        self.y_ = y

        self._init_metrics()

        super().fit(X, y, grouping = grouping, abs_val = abs_val, gamma = gamma, protected_features=protected_features, Xp=Xp, **kwargs)

        # Return the classifier
        return self

    def predict(self, X):
        """ Predict labels from X.

        Parameters
        ----------
        X : array-like, shape (n_samples, n_features)
            The input samples.

        Returns
        -------
        y : ndarray, shape (n_samples,)
            The label for each sample is the label of the closest sample
            seen during fit.
        """
        # Check is fit had been called
        check_is_fitted(self, ['X_', 'y_'])

        # Input validation
        # X = check_array(X)

        return super().predict(X)

    def predict_proba(self, X):
        """ Return prediction probabilities."""
        return self.best_estimator_.predict_proba(X)

    def _init_metrics(self):
        """ Check metric definitions and/or define when necessary. """
        self.accuracy_metrics_ = self.accuracy_metrics
        self.fairness_metrics_ = self.fairness_metrics
        if self.accuracy_metrics is None:
            self.accuracy_metrics_ = [make_scorer(roc_auc_score, greater_is_better=False, needs_proba=True)]
        if self.fairness_metrics is None:
            self.fairness_metrics_ = [metrics.multicalibration_loss]

    def predict_proba_archive(self, X):
        """Return a list of predictions from the archive models. """
        check_is_fitted(self, 'is_fitted_')
        if not hasattr(self, 'estimator_archive_'):
            print("Need to fit archive models. You can set `store_final_models` to True to avoid this step.")
            self.estimator_archive_ = self._store_final_models()
        return self._output_archive('predict_proba', X=X)


class FomoRegressor(RegressorMixin, BaseEstimator):
    """Fomo class for regression models. 

    Parameters
    ----------
    estimator : sklearn-like estimator
        The underlying ML model to be trained. 
        The ML model must accept `sample_weight` as an argument to :meth:`fit`. 
    fairness_metrics : list[Callable]
        The fairness metrics to try to optimize during fitting. 
    accuracy_metrics : list[Callable]
        The accuracy metrics to try to optimize during fitting. 
    algorithm: pymoo Algorithm
        The multi-objective optimizer to use. Should be compatible with 
        `pymoo.core.algorithm.Algorithm`. 
    random_state: int | None
        Random seed. 
    verbose: bool
        Whether to print progress.
    n_jobs: int
        Number of parallel processes to use. Parallelizes evaluation.
    store_final_models: bool
        If True, the final set of models will be stored in the estimator.
    problem_type: ElementwiseProblem
        Determines the evaluation class to be used. Options:

        - :class:`BasicProblem`: loss function weights are directly optimized.
        - :class:`MLPProblem`: weights of a multilayer perceptron are optimized to estimate loss function weights. 
        - :class:`LinearProblem`: weights of a logistic model are optimized to estimate loss function weights.

    Attributes
    ----------
    n_features_ : int
        The number of features of the data passed to :meth:`fit`.

    Examples
    --------
    >>> from fomo import FomoRegressor
    >>> from pmlb import pmlb
    >>> X,y = pmlb.fetch_data('adult', return_X_y=True)
    >>> groups = ['race','sex']
    >>> est = FomoRegressor()
    >>> est.fit(X,y, protected_features=groups)
    """
    def __init__(self, 
                 estimator: RegressorMixin=SGDRegressor(),
                 fairness_metrics: list[str]=None,
                 accuracy_metrics: list[str]=None,
                 algorithm: str ='NSGA2',
                 random_state: int=None,
                 verbose: bool = False,
                 n_jobs: int = -1,
                 store_final_models: bool = False,
                 problem_type = BasicProblem,
                 checkpoint:bool = False,
                 picking_strategy: str = 'PseudoWeights'
                ):
        super().__init__(
            estimator, 
            fairness_metrics, 
            accuracy_metrics,
            algorithm, 
            random_state,
            verbose,
            n_jobs,
            store_final_models,
            problem_type,
            checkpoint,
            picking_strategy
        )

    def fit(self, X, y, protected_features=None, Xp=None, **kwargs):
        """Train a set of regressors. 

        Parameters
        ----------
        X : array-like, shape (n_samples, n_features)
            The training input samples.
        y : array-like, shape (n_samples,)
            The target values. An array of int.
        protected_features: list|None, default = None
            The columns of X to calculate fairness over. If specifying columns,
            do not also specify `Xp`.
        Xp: pandas DataFrame, shape (n_samples, n_protected_features), default=None
            The protected feature values used to calculate fairness. If `Xp` is 
            specified, `protected_features` must be None. 

        Returns
        -------
        self : object
            Returns self.
        """
        # Check that X and y have correct shape
        # X, y = check_X_y(X, y)
        # Store the classes seen during fit
        self.classes_ = unique_labels(y)

        self.X_ = X
        self.y_ = y
        super().fit(X, y, protected_features=protected_features, Xp=Xp, **kwargs)

        # Return the regressor
        return self

    def predict(self, X):
        """Predict outcome. 

        Parameters
        ----------
        X : array-like, shape (n_samples, n_features)
            The input samples.

        Returns
        -------
        y : ndarray, shape (n_samples,)
            The label for each sample is the label of the closest sample
            seen during fit.
        """
        # Check is fit had been called
        check_is_fitted(self, ['X_', 'y_'])

        # Input validation
        # X = check_array(X)

        return super().predict(X)

    def _init_metrics(self):
        """ Check metric definitions and/or define when necessary. """
        self.accuracy_metrics_ = []
        self.fairness_metrics_ = []
        if self.accuracy_metrics is None:
            # self.accuracy_metrics_.append(make_scorer(r2_score, greater_is_better=False))
            self.accuracy_metrics_.append(make_scorer(mean_squared_error))
        if len(self.fairness_metrics) == 0:
            self.fairness_metrics_.append(metrics.subgroup_MSE)
